@@ -18,7 +18,7 @@ from itertools import cycle
 from pathlib import Path
 from threading import Thread
 from typing import List, Dict, Any, Tuple
-from xml.etree import ElementTree
+from xml.etree import ElementTree as ET
 from xml.etree.ElementTree import ParseError, Element
 
 import aiohttp
@@ -91,6 +91,10 @@ class Deskshare:
     height: int
 
 
+class ContentRangeError(ConnectionError):
+    pass
+
+
 class BBBDL:
     VALID_URL_RE = re.compile(
         r'''(?x)
@@ -186,7 +190,18 @@ class BBBDL:
         _ = asyncio.run(self.batch_download_from_bbb(dl_jobs))
 
         Log.info("Downloading webcams / deskshare")
-        dl_jobs = ['cursor.xml', 'panzooms.xml', 'captions.json', 'deskshare.xml', 'events.xml']
+        dl_jobs = [
+            'cursor.xml',
+            'panzooms.xml',
+            'captions.json',
+            'deskshare.xml',
+            'events.xml',
+            'presentation_text.json',
+            'slides_new.xml',
+            'notes.html',
+            'polls.json',
+            'external_videos.json',
+        ]
         cam_webm_idx = append_get_idx(dl_jobs, 'video/webcams.webm')
         cam_mp4_idx = append_get_idx(dl_jobs, 'video/webcams.mp4')
         dsk_webm_idx = append_get_idx(dl_jobs, 'deskshare/deskshare.webm')
@@ -255,6 +270,41 @@ class BBBDL:
             Log.warning(f'Temporary directory will not be deleted: {self.tmp_dir}')
         Log.success(f'All done! Final video: {result_path}')
 
+    def run_audio_only(self):
+        if not self.backup:
+            Log.yellow(f'Output directory for the final audio is: {self.output_dir}')
+            Log.yellow(f'Directory for the temporary files is: {self.tmp_dir}')
+        else:
+            Log.error('Please use the backup option only without the audio only mode')
+            exit(-11)
+
+        Log.info("Downloading meta information")
+
+        dl_jobs = ['metadata.xml']
+        _ = asyncio.run(self.batch_download_from_bbb(dl_jobs))
+
+        Log.info("Downloading webcams file")
+        dl_jobs = []
+        cam_webm_idx = append_get_idx(dl_jobs, 'video/webcams.webm')
+        cam_mp4_idx = append_get_idx(dl_jobs, 'video/webcams.mp4')
+
+        dl_results = asyncio.run(self.batch_download_from_bbb(dl_jobs, False))
+
+        if not dl_results[cam_webm_idx] and not dl_results[cam_mp4_idx]:
+            Log.error('Error: webcams video is essential. Abort! Please try again later!')
+            exit(4)
+        webcams_rel_path = 'video/webcams.webm' if dl_results[cam_webm_idx] else 'video/webcams.mp4'
+        webcams_path = PT.get_in_dir(self.tmp_dir, webcams_rel_path)
+
+        metadata = self.parse_metadata()
+        result_path = self.extract_audio(webcams_path, metadata)
+
+        if not self.keep_tmp_files:
+            self.remove_tmp_dir()
+        else:
+            Log.warning(f'Temporary directory will not be deleted: {self.tmp_dir}')
+        Log.success(f'All done! Final audio: {result_path}')
+
     def parse_deskshare_data(self, recording_duration) -> List[Deskshare]:
         result_list = []
         loaded_deskshare = self.load_xml('deskshare.xml', False)
@@ -361,15 +411,17 @@ class BBBDL:
         status_dict: Dict,
     ):
         semaphore = asyncio.Semaphore(self.max_parallel_chromes)
+        gather_jobs = asyncio.gather(
+            *[
+                self.capture_frames(server_url, frames, only_zooms, partition, semaphore, status_dict)
+                for partition in partitions
+            ]
+        )
         try:
-            await asyncio.gather(
-                *[
-                    self.capture_frames(server_url, frames, only_zooms, partition, semaphore, status_dict)
-                    for partition in partitions
-                ]
-            )
+            await gather_jobs
         except Exception:
             traceback.print_exc()
+            gather_jobs.cancel()
             Log.error(
                 'Unexpected Error! Press Ctr+C to exit.'
                 + ' Please try to set a low number of threads with `--max-parallel-chromes`.'
@@ -796,6 +848,14 @@ class BBBDL:
                 Path(self.output_dir) / PT.to_valid_name(metadata.date_formatted + '_' + metadata.title + '.mp4')
             )
 
+    def get_output_audio_file_path(self, metadata: Metadata):
+        if self.filename is not None:
+            return str(Path(self.output_dir) / PT.to_valid_name(self.filename))
+        else:
+            return str(
+                Path(self.output_dir) / PT.to_valid_name(metadata.date_formatted + '_' + metadata.title + '.mp3')
+            )
+
     def get_output_dir(
         self,
         output_dir: str,
@@ -868,20 +928,6 @@ class BBBDL:
         assert not rel_file_path.startswith('/') and not rel_file_path.startswith('\\')
         return self.presentation_base_url + '/' + rel_file_path
 
-    async def readexactly(self, steam, n):
-        if steam._exception is not None:
-            raise steam._exception
-
-        blocks = []
-        while n > 0:
-            block = await steam.read(n)
-            if not block:
-                break
-            blocks.append(block)
-            n -= len(block)
-
-        return b''.join(blocks)
-
     async def get_can_continue_on_fail(self, url, session):
         try:
             headers = self.headers.copy()
@@ -945,11 +991,12 @@ class BBBDL:
                     try:
                         if tries_num > 0 and can_continue_on_fail:
                             headers["Range"] = f"bytes={received}-"
+                        elif not can_continue_on_fail and 'Range' in headers:
+                            del headers['Range']
                         ssl_param = False if self.no_check_certificate else None
                         async with session.request(
                             "GET", dl_url, headers=headers, raise_for_status=True, ssl=ssl_param
                         ) as resp:
-
                             # Download the file.
                             total = int(resp.headers.get("Content-Length", 0))
                             content_range = resp.headers.get("Content-Range", "")  # Example: bytes 200-1000/67589
@@ -959,18 +1006,18 @@ class BBBDL:
                                     Log.debug(f"Warning {rel_file_path} got status {resp.status}")
 
                             if tries_num > 0 and can_continue_on_fail and not content_range and resp.status != 206:
-                                raise ClientError(
+                                raise ContentRangeError(
                                     f"Server did not response for {rel_file_path} with requested range data"
                                 )
                             file_obj = file_obj or await aiofiles.open(local_path, "wb")
-                            chunk = await self.readexactly(resp.content, 1024000)
+                            chunk = await resp.content.read(1024 * 10)
                             chunk_idx = 0
                             while chunk:
                                 received += len(chunk)
-                                if chunk_idx % 10 == 0:
+                                if chunk_idx % 100 == 0:
                                     Log.info(f"{rel_file_path} got {format_bytes(received)} / {format_bytes(total)}")
                                 await file_obj.write(chunk)
-                                chunk = await self.readexactly(resp.content, 1024000)
+                                chunk = await resp.content.read(1024 * 10)
                                 chunk_idx += 1
 
                         if self.verbose:
@@ -981,21 +1028,29 @@ class BBBDL:
                         finished_successfully = True
                         break
 
-                    except (ClientError, OSError, ValueError) as err:
+                    except (ClientError, OSError, ValueError, ContentRangeError) as err:
                         if tries_num == 0:
                             can_continue_on_fail = await self.get_can_continue_on_fail(dl_url, session)
-                        if not can_continue_on_fail:
+                        if (not can_continue_on_fail and received > 0) or isinstance(err, ContentRangeError):
+                            can_continue_on_fail = False
                             # Clean up failed file because we can not recover
                             if file_obj is not None:
-                                file_obj.close()
+                                await file_obj.close()
+                                file_obj = None
                             if os.path.exists(local_path):
                                 os.unlink(local_path)
+                            received = 0
 
-                        if isinstance(err, ClientResponseError) and err.status == 404 and not is_essential:
-                            Log.info(f'{rel_file_path} could not be downloaded: {err.status} {err.message}')
-                            if self.verbose:
-                                Log.info(f'Error: {str(err)}')
-                            break
+                        if isinstance(err, ClientResponseError):
+                            if err.status in [408, 409, 429]:   # pylint: disable=no-member
+                                # 408 (timeout) or 409 (conflict) and 429 (too many requests)
+                                # Retry after 1 sec
+                                await asyncio.sleep(1)
+                            else:
+                                Log.info(f'{rel_file_path} could not be downloaded: {err.status} {err.message}')
+                                if self.verbose:
+                                    Log.info(f'Error: {str(err)}')
+                                break
 
                         if self.verbose:
                             Log.warning(
@@ -1005,8 +1060,10 @@ class BBBDL:
                         tries_num += 1
 
             if file_obj is not None:
-                file_obj.close()
+                await file_obj.close()
             if not finished_successfully:
+                if os.path.exists(local_path):
+                    os.unlink(local_path)
                 return False
             return True
 
@@ -1014,7 +1071,7 @@ class BBBDL:
         local_path = PT.get_in_dir(self.tmp_dir, rel_file_path)
         if os.path.exists(local_path):
             try:
-                tree_root = ElementTree.parse(local_path).getroot()
+                tree_root = ET.parse(local_path).getroot()
                 return tree_root
             except ParseError as err:
                 Log.error(f'Unable to parse XML file "{local_path}": {str(err)}')
@@ -1037,7 +1094,6 @@ class BBBDL:
         webcams_rel_path: str,
         metadata: Metadata,
     ):
-
         webcam_is_empty = False
         if not self.skip_webcam_opt and not self.skip_webcam_freeze_detection_opt:
             Log.info(f'Try to detect freeze in {webcams_rel_path}...')
@@ -1172,6 +1228,21 @@ class BBBDL:
         Log.info(f'Creating slideshow finished and took: {formatSeconds(t.duration)}')
         return slideshow_path
 
+    def extract_audio(
+        self,
+        webcams_path: str,
+        metadata: Metadata,
+    ):
+        Log.info('Start extracting audio...')
+        result_path = self.get_output_audio_file_path(metadata)
+        if os.path.isfile(result_path):
+            Log.warning('Final Audio already exists. Abort!')
+            return result_path
+        with Timer() as t:
+            asyncio.run(self.ffmpeg.extract_audio(webcams_path, result_path))
+        Log.info(f'Extracting audio finished and took: {formatSeconds(t.duration)}')
+        return result_path
+
 
 def get_parser():
     """
@@ -1183,6 +1254,13 @@ def get_parser():
 
     parser.add_argument('URL', type=str, help='URL of a BBB lesson')
 
+    parser.add_argument(
+        '-ao',
+        '--audio-only',
+        action='store_true',
+        help='Extract only the audio from the presentation, do not generate video.',
+    )
+
     parser.add_argument(
         '-sw',
         '--skip-webcam',
@@ -1358,7 +1436,7 @@ def main(args=None):
     args = parser.parse_args(args)
 
     with Timer() as final_t:
-        BBBDL(
+        bbb_dl = BBBDL(
             args.URL,
             args.filename,
             args.output_dir,
@@ -1380,5 +1458,9 @@ def main(args=None):
             args.force_height,
             args.preset,
             args.crf,
-        ).run()
+        )
+        if args.audio_only:
+            bbb_dl.run_audio_only()
+        else:
+            bbb_dl.run()
     Log.info(f'BBB-DL finished and took: {formatSeconds(final_t.duration)}')
